@@ -47,15 +47,20 @@ from molecule.scenario import Scenario
 
 if TYPE_CHECKING:
     from collections.abc import Callable
+    from typing import NoReturn
 
     from molecule.scenario import Scenario
+    from molecule.types import CommandArgs, MoleculeArgs
+
+    ClickCommand = Callable[[Callable[..., None]], click.Command]
+    ClickGroup = Callable[[Callable[..., None]], click.Group]
 
 LOG = logging.getLogger(__name__)
 MOLECULE_GLOB = os.environ.get("MOLECULE_GLOB", "molecule/*/molecule.yml")
 MOLECULE_DEFAULT_SCENARIO_NAME = "default"
 
 
-class Base(metaclass=abc.ABCMeta):
+class Base(abc.ABC):
     """An abstract base class used to define the command interface."""
 
     def __init__(self, c: config.Config) -> None:
@@ -63,9 +68,6 @@ class Base(metaclass=abc.ABCMeta):
 
         Args:
             c: An instance of a Molecule config.
-
-        Returns:
-            None
         """
         self._config = c
         self._setup()
@@ -74,11 +76,11 @@ class Base(metaclass=abc.ABCMeta):
         """Decorate execute from all subclasses."""
         super().__init_subclass__()
         for wrapper in logger.get_section_loggers():
-            cls.execute = wrapper(cls.execute)  # type: ignore  # noqa: PGH003
+            cls.execute = wrapper(cls.execute)  # type: ignore[method-assign]
 
     @abc.abstractmethod
     def execute(
-        self: Base,
+        self,
         action_args: list[str] | None = None,
     ) -> None:  # pragma: no cover
         """Abstract method to execute the command.
@@ -90,15 +92,16 @@ class Base(metaclass=abc.ABCMeta):
     def _setup(self) -> None:
         """Prepare Molecule's provisioner and returns None."""
         self._config.write()
-        self._config.provisioner.write_config()
-        self._config.provisioner.manage_inventory()
+        if self._config.provisioner is not None:
+            self._config.provisioner.write_config()
+            self._config.provisioner.manage_inventory()
 
 
 def execute_cmdline_scenarios(
     scenario_name: str | None,
-    args: dict[str, Any],
-    command_args: dict[str, Any],
-    ansible_args: tuple[()] | tuple[str] = (),
+    args: MoleculeArgs,
+    command_args: CommandArgs,
+    ansible_args: tuple[str, ...] = (),
 ) -> None:
     """Execute scenario sequences based on parsed command-line arguments.
 
@@ -113,12 +116,15 @@ def execute_cmdline_scenarios(
         args: ``args`` dict from ``click`` command context
         command_args: dict of command arguments, including the target
         ansible_args: Optional tuple of arguments to pass to the `ansible-playbook` command
+
+    Raises:
+        SystemExit: If scenario exits prematurely.
     """
     glob_str = MOLECULE_GLOB
     if scenario_name:
         glob_str = glob_str.replace("*", scenario_name)
     scenarios = molecule.scenarios.Scenarios(
-        get_configs(args, command_args, ansible_args, glob_str),  # type: ignore[no-untyped-call]
+        get_configs(args, command_args, ansible_args, glob_str),
         scenario_name,
     )
 
@@ -173,10 +179,13 @@ def execute_subcommand(
     Args:
         current_config: An instance of a Molecule config.
         subcommand_and_args: A string representing the subcommand and arguments.
+
+    Returns:
+        The result of the subcommand.
     """
     (subcommand, *args) = subcommand_and_args.split(" ")
     command_module = getattr(molecule.command, subcommand)
-    command = getattr(command_module, text.camelize(subcommand))  # type: ignore[no-untyped-call]
+    command = getattr(command_module, text.camelize(subcommand))
 
     # knowledge of the current action is used by some provisioners
     # to ensure they behave correctly during certain sequence steps,
@@ -203,7 +212,15 @@ def execute_scenario(scenario: Scenario) -> None:
             scenario._remove_scenario_state_directory()  # noqa: SLF001
 
 
-def filter_ignored_scenarios(scenario_paths) -> list[str]:  # type: ignore[no-untyped-def]  # noqa: ANN001, D103
+def filter_ignored_scenarios(scenario_paths: list[str]) -> list[str]:
+    """Filter out candidate scenario paths that are ignored by git.
+
+    Args:
+        scenario_paths: List of candidate scenario paths.
+
+    Returns:
+        Filtered list of scenario paths.
+    """
     command = ["git", "check-ignore", *scenario_paths]
 
     with contextlib.suppress(subprocess.CalledProcessError, FileNotFoundError):
@@ -224,7 +241,12 @@ def filter_ignored_scenarios(scenario_paths) -> list[str]:  # type: ignore[no-un
     return paths
 
 
-def get_configs(args, command_args, ansible_args=(), glob_str=MOLECULE_GLOB):  # type: ignore[no-untyped-def]  # noqa: ANN001, ANN201
+def get_configs(
+    args: MoleculeArgs,
+    command_args: CommandArgs,
+    ansible_args: tuple[str, ...] = (),
+    glob_str: str = MOLECULE_GLOB,
+) -> list[config.Config]:
     """Glob the current directory for Molecule config files.
 
     Instantiate config objects, and returns a list.
@@ -246,7 +268,7 @@ def get_configs(args, command_args, ansible_args=(), glob_str=MOLECULE_GLOB):  #
     scenario_paths = filter_ignored_scenarios(scenario_paths)
     configs = [
         config.Config(
-            molecule_file=util.abs_path(c),  # type: ignore[arg-type]
+            molecule_file=util.abs_path(c),
             args=args,
             command_args=command_args,
             ansible_args=ansible_args,
@@ -289,8 +311,12 @@ def _get_subcommand(string: str) -> str:
     return string.split(".")[-1]
 
 
-def click_group_ex():  # type: ignore[no-untyped-def]  # noqa: ANN201
-    """Return extended version of click.group()."""
+def click_group_ex() -> ClickGroup:
+    """Return extended version of click.group().
+
+    Returns:
+        Click command group.
+    """
     # Color coding used to group command types, documented only here as we may
     # decide to change them later.
     # green : (default) as sequence step
@@ -300,7 +326,11 @@ def click_group_ex():  # type: ignore[no-untyped-def]  # noqa: ANN201
         cls=HelpColorsGroup,
         # Workaround to disable click help line truncation to ~80 chars
         # https://github.com/pallets/click/issues/486
-        context_settings={"max_content_width": 9999, "color": should_do_markup()},
+        context_settings={
+            "max_content_width": 9999,
+            "color": should_do_markup(),
+            "help_option_names": ["-h", "--help"],
+        },
         help_headers_color="yellow",
         help_options_color="green",
         help_options_custom_colors={
@@ -316,17 +346,33 @@ def click_group_ex():  # type: ignore[no-untyped-def]  # noqa: ANN201
     )
 
 
-def click_command_ex() -> Callable[[Callable[..., Any]], click.Command]:
-    """Return extended version of click.command()."""
+def click_command_ex(name: str | None = None) -> ClickCommand:
+    """Return extended version of click.command().
+
+    Args:
+        name: A replacement name in the case the automatic one is insufficient.
+
+    Returns:
+        Click command group.
+    """
     return click.command(
         cls=HelpColorsCommand,
+        name=name,
         help_headers_color="yellow",
         help_options_color="green",
     )
 
 
-def result_callback(*args, **kwargs):  # type: ignore[no-untyped-def]  # noqa: ANN002, ANN003, ANN201, ARG001
-    """Click natural exit callback."""
+def result_callback(
+    *args: object,  # noqa: ARG001
+    **kwargs: object,  # noqa: ARG001
+) -> NoReturn:
+    """Click natural exit callback.
+
+    Args:
+        *args: Unused.
+        **kwargs: Unused.
+    """
     # We want to be used we run out custom exit code, regardless if run was
     # a success or failure.
     util.sysexit(0)
